# -*- coding: utf-8 -*-

"""
SPDX-FileCopyrightText: Deutsches Zentrum für Luft und Raumfahrt
SPDX-FileCopyrightText: Patrik Schönfeldt

SPDX-License-Identifier: MIT
"""

__version__ = "2.1.0rc3"

from . import layered_heat
from . import physics
from .meta_model import MetaModel
